@@ -25,6 +25,7 @@ import shutil
 from learning_loop_node.data_classes.category import Category
 from learning_loop_node.trainer.hyperparameter import Hyperparameter
 import time
+from time import perf_counter
 
 
 class Trainer():
@@ -41,6 +42,7 @@ class Trainer():
         self.training.data = await downloader.download_training_data(self.training.images_folder)
         self.training.data.categories = Category.from_list(details['categories'])
         self.training.data.hyperparameter = Hyperparameter.from_dict(details)
+        self.training.training_number = details['training_number']
 
         base_model_id = details['id']
         self.training.base_model_id = base_model_id
@@ -89,8 +91,8 @@ class Trainer():
     def get_log(self) -> str:
         return self.executor.get_log()
 
-    async def save_model(self,  context: Context, model_id: str) -> None:
-        files = await asyncio.get_running_loop().run_in_executor(None, self.get_model_files, model_id)
+    async def save_model(self, context: Context) -> dict:
+        files = await asyncio.get_running_loop().run_in_executor(None, self.get_latest_model_files)
         model_json_content = self.create_model_json_content()
         model_json_path = '/tmp/model.json'
         with open(model_json_path, 'w') as f:
@@ -98,7 +100,7 @@ class Trainer():
 
         if isinstance(files, list):
             files = {self.model_format: files}
-
+        uploaded_model = None
         if isinstance(files, dict):
             for format in files:
                 # model.json was mandatory in previous versions. Now its forbidden to provide an own model.json file.
@@ -106,9 +108,11 @@ class Trainer():
                     "It is not allowed to provide a 'model.json' file."
                 _files = files[format]
                 _files.append(model_json_path)
-                await uploads.upload_model(context, _files, model_id, format)
+                uploaded_model = await uploads.upload_model_for_training(context, _files, self.training.training_number, format)
+
         else:
             raise TypeError(f'can only save model as list or dict, but was {files}')
+        return uploaded_model
 
     def get_new_model(self) -> Optional[BasicModel]:
         '''Is called frequently to check if a new "best" model is availabe.
@@ -120,16 +124,16 @@ class Trainer():
         '''
         raise NotImplementedError()
 
-    def on_model_published(self, basic_model: BasicModel, model_id: str) -> None:
+    def on_model_published(self, basic_model: BasicModel) -> None:
         '''Called after a BasicModel has been successfully send to the Learning Loop.
-        The model_id is an uuid to identify the model within the Learning Loop.
-        self.get_model_files uses this id to gather all files needed for transfering the actual data from the trainer node to the Learning Loop.
-        In the simplest implementation this method just renames the weight file (encoded in BasicModel.meta_information) into a file name containing the model_id.
+        The files for this model should be stored.
+        self.get_latest_model_files is used to gather all files needed for transfering the actual data from the trainer node to the Learning Loop.
+        In the simplest implementation this method just renames the weight file (encoded in BasicModel.meta_information) into a file name like latest_published_model
         '''
         raise NotImplementedError()
 
-    def get_model_files(self, model_id: str) -> Union[List[str], Dict[str, List[str]]]:
-        '''Called when the Learning Loop requests to backup a specific model. 
+    def get_latest_model_files(self) -> Union[List[str], Dict[str, List[str]]]:
+        '''Called when the Learning Loop requests to backup the latest model for the training. 
         Should return a list of file paths which describe the model.
         These files must contain all data neccessary for the trainer to resume a training (eg. weight file, hyperparameters, etc.) 
         and will be stored in the Learning Loop unter the format of this trainer.
@@ -141,16 +145,18 @@ class Trainer():
         '''
         raise NotImplementedError()
 
-    async def do_detections(self, context: Context, model_id: str, model_format: str):
-        tmp_folder = f'/tmp/model_for_auto_detections_{model_id}_{model_format}'
+    async def do_detections(self, context: Context, model_id: str):
+
+        tmp_folder = f'/tmp/model_for_auto_detections_{model_id}_{self.model_format}'
 
         shutil.rmtree(tmp_folder, ignore_errors=True)
         os.makedirs(tmp_folder)
         logging.info('downloading model for detecting')
         try:
-            await downloads.download_model(tmp_folder, context, model_id, model_format)
+            await downloads.download_model(tmp_folder, context, model_id, self.model_format)
         except:
             logging.exception('download error')
+            return
         with open(f'{tmp_folder}/model.json', 'r') as f:
             content = json.load(f)
             model_information = ModelInformation.parse_obj(content)
@@ -160,17 +166,28 @@ class Trainer():
         downloader = DataDownloader(context)
         image_ids = []
         for state in ['inbox', 'annotate', 'review', 'complete']:
+            logging.info(f'fetching image ids of {state}')
             new_ids = await downloader.fetch_image_ids(query_params=f'state={state}')
             image_ids += new_ids
+            logging.info(f'downloading {len(new_ids)} images')
             await downloader.download_images(new_ids, image_folder)
-        images = [img for img in glob(f'{image_folder}/**/*.*', recursive=True)
-                  if os.path.splitext(os.path.basename(img))[0] in image_ids]
+        images = await asyncio.get_event_loop().run_in_executor(None, Trainer.images_for_ids, image_ids, image_folder)
         logging.info(f'running detections on {len(images)} images')
         detections = await self._detect(model_information, images, tmp_folder)
         logging.info(f'uploading {len(detections)} detections')
         await self._upload_detections(context, jsonable_encoder(detections))
         return detections
 
+    @staticmethod
+    def images_for_ids(image_ids, image_folder) -> List[str]:
+        logging.info(f'### Going to get images for {len(image_ids)} images ids')
+        start = perf_counter()
+        images = [img for img in glob(f'{image_folder}/**/*.*', recursive=True)
+                  if os.path.splitext(os.path.basename(img))[0] in image_ids]
+        end = perf_counter()
+        logging.info(f'found {len(images)} images for {len(image_ids)} image ids, which took {end-start:0.2f} seconds')
+        return images
+
     async def _detect(self, model_information: ModelInformation, images:  List[str], model_folder: str) -> List:
         raise NotImplementedError()
 
@@ -179,10 +196,9 @@ class Trainer():
         batch_size = 500
         for i in tqdm(range(0, len(detections), batch_size), position=0, leave=True):
             batch_detections = detections[i:i+batch_size]
-            data = json.dumps(batch_detections)
-            logging.info(f'uploading detections. File size : {len(data)}')
+            logging.info(f'uploading detections. File size : {len(json.dumps(batch_detections))}')
             try:
-                async with loop.post(f'api/{context.organization}/projects/{context.project}/detections', data=data) as response:
+                async with loop.post(f'api/{context.organization}/projects/{context.project}/detections', json=batch_detections) as response:
                     if response.status != 200:
                         logging.error(f'could not upload detections. {str(response)}')
                     else:
@@ -220,12 +236,12 @@ class Trainer():
         return training_folder
 
     @property
-    def hyperparameters(self) -> str:
+    def hyperparameters(self) -> dict:
         if self.training and self.training.data:
-            information = f"resolution: {self.training.data.hyperparameter.resolution}"
-            information += f"\nflip right/left: {self.training.data.hyperparameter.flip_rl}"
-            information += f"\nflip up/down: {self.training.data.hyperparameter.flip_ud}"
-            information += f"\nmodel architecture: {self.model_architecture}" if self.model_architecture else ""
+            information = {}
+            information['resolution'] = self.training.data.hyperparameter.resolution
+            information['flipRl'] = self.training.data.hyperparameter.flip_rl
+            information['flipUd'] = self.training.data.hyperparameter.flip_ud
             return information
         else:
             return None
